import logging
import os
import socket
import sys
import threading
import time
from unittest.mock import Mock, patch

import pytest
from vine import promise

from celery.backends.asynchronous import E_CELERY_RESTART_REQUIRED, BaseResultConsumer, greenletDrainer
from celery.backends.base import Backend
from celery.utils import cached_property

# ---- helpers ---------------------------------------------------------------


def _make_consumer(app, environment='default'):
    """Create a BaseResultConsumer with a mocked drainer environment."""
    with patch('celery.backends.asynchronous.detect_environment') as det:
        det.return_value = environment
        backend = Backend(app)
        consumer = BaseResultConsumer(
            backend, app, backend.accept,
            pending_results={}, pending_messages={},
        )
    return consumer


# ---------------------------------------------------------------------------
# 1. Drainer (default / synchronous) -- no gevent / eventlet needed
# ---------------------------------------------------------------------------

class test_Drainer_without_greenlets:

    # -- drain_events_until: normal flow ------------------------------------

    def test_drain_fulfils_promise(self, app):
        """Loop exits once the promise is fulfilled."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        calls = [0]

        def wait(timeout=None):
            calls[0] += 1
            if calls[0] >= 3:
                p('done')

        for _ in drainer.drain_events_until(
                p, wait=wait, interval=0.01, timeout=5):
            pass

        assert p.ready
        assert calls[0] >= 3

    def test_drain_calls_on_interval(self, app):
        """on_interval callback is invoked every iteration."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        on_interval = Mock()
        calls = [0]

        def wait(timeout=None):
            calls[0] += 1
            if calls[0] >= 3:
                p('done')

        for _ in drainer.drain_events_until(
                p, wait=wait, interval=0.01, timeout=5,
                on_interval=on_interval):
            pass

        assert on_interval.call_count >= 2

    def test_drain_raises_timeout(self, app):
        """socket.timeout raised when total elapsed time exceeds *timeout*."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()

        def wait(timeout=None):
            time.sleep(0.02)

        with pytest.raises(socket.timeout):
            for _ in drainer.drain_events_until(
                    p, wait=wait, interval=0.01, timeout=0.05):
                pass

        assert not p.ready

    def test_drain_uses_result_consumer_drain_events_by_default(self, app):
        """When *wait* is None, result_consumer.drain_events is used."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        calls = [0]

        def mock_drain(timeout=None):
            calls[0] += 1
            if calls[0] >= 2:
                p('done')

        consumer.drain_events = mock_drain

        for _ in drainer.drain_events_until(p, interval=0.01, timeout=5):
            pass

        assert p.ready
        assert calls[0] >= 2

    # -- drain_events_until: socket.timeout from wait -----------------------

    def test_drain_swallows_socket_timeout_from_wait(self, app):
        """socket.timeout raised inside wait() must be silently caught."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        calls = [0]

        def wait(timeout=None):
            calls[0] += 1
            if calls[0] <= 2:
                raise socket.timeout('idle')
            p('done')

        for _ in drainer.drain_events_until(
                p, wait=wait, interval=0.01, timeout=5):
            pass

        assert p.ready

    # -- drain_events_until: OSError from wait ------------------------------

    def test_drain_catches_oserror_and_logs(self, app):
        """OSError from wait() must be caught, logged, loop continues."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        calls = [0]

        def wait(timeout=None):
            calls[0] += 1
            if calls[0] <= 2:
                raise OSError('broker away')
            p('done')

        with patch.object(logging, 'warning') as mock_warn:
            for _ in drainer.drain_events_until(
                    p, wait=wait, interval=0.01, timeout=5):
                pass

        assert p.ready
        assert mock_warn.call_count >= 2

    # -- wait_for -----------------------------------------------------------

    def test_wait_for_calls_wait_with_timeout(self, app):
        """Drainer.wait_for delegates to the wait callback."""
        consumer = _make_consumer(app)
        drainer = consumer.drainer
        p = promise()
        wait = Mock()
        drainer.wait_for(p, wait, timeout=0.5)
        wait.assert_called_once_with(timeout=0.5)


# ---------------------------------------------------------------------------
# 2. greenletDrainer -- tested synchronously (no real greenlet spawning)
# ---------------------------------------------------------------------------

class test_greenletDrainer:

    def _make_greenlet_drainer(self, app):
        consumer = _make_consumer(app)
        drainer = greenletDrainer(consumer)
        return drainer

    # -- run: normal stop ---------------------------------------------------

    def test_run_exits_when_stopped(self, app):
        """run() exits cleanly when _stopped is set."""
        drainer = self._make_greenlet_drainer(app)
        calls = [0]

        def drain(timeout=None):
            calls[0] += 1
            if calls[0] >= 3:
                drainer._stopped.set()

        drainer.result_consumer.drain_events = Mock(side_effect=drain)
        drainer.run()

        assert drainer._shutdown.is_set()
        assert drainer._exc is None

    # -- run: socket.timeout is swallowed -----------------------------------

    def test_run_swallows_socket_timeout(self, app):
        """socket.timeout inside run() must be silently caught."""
        drainer = self._make_greenlet_drainer(app)
        calls = [0]

        def drain(timeout=None):
            calls[0] += 1
            if calls[0] <= 3:
                raise socket.timeout('idle')
            drainer._stopped.set()

        drainer.result_consumer.drain_events = Mock(side_effect=drain)
        drainer.run()

        assert calls[0] >= 4
        assert drainer._exc is None

    # -- run: OSError is caught and logged ----------------------------------

    def test_run_catches_oserror_and_logs(self, app):
        """OSError in run() must be caught/logged, loop continues."""
        drainer = self._make_greenlet_drainer(app)
        calls = [0]

        def drain(timeout=None):
            calls[0] += 1
            if calls[0] <= 3:
                raise OSError('connection reset')
            drainer._stopped.set()

        drainer.result_consumer.drain_events = Mock(side_effect=drain)

        with patch.object(logging, 'warning') as mock_warn, \
                patch('celery.backends.asynchronous.time.sleep') as mock_sleep:
            drainer.run()

        assert calls[0] >= 4
        assert mock_warn.call_count >= 3
        # backoff sleep should have been called once per OSError
        assert mock_sleep.call_count >= 3
        assert drainer._exc is None

    # -- run: unexpected Exception is stored and re-raised ------------------

    def test_run_stores_and_reraises_unexpected_exception(self, app):
        """Non-OSError / non-timeout exceptions must propagate and be stored."""
        drainer = self._make_greenlet_drainer(app)

        def drain(timeout=None):
            raise RuntimeError('unexpected')

        drainer.result_consumer.drain_events = Mock(side_effect=drain)

        with pytest.raises(RuntimeError, match='unexpected'):
            drainer.run()

        assert drainer._exc is not None
        assert drainer._shutdown.is_set()

    # -- _ensure_not_shut_down ----------------------------------------------

    def test_ensure_not_shut_down_raises_stored_exc(self, app):
        """If run() failed, _ensure_not_shut_down re-raises the exception."""
        drainer = self._make_greenlet_drainer(app)
        drainer._shutdown.set()
        drainer._exc = ValueError('boom')

        with pytest.raises(ValueError, match='boom'):
            drainer._ensure_not_shut_down()

    def test_ensure_not_shut_down_raises_restart_msg(self, app):
        """If stopped cleanly, _ensure_not_shut_down raises restart msg."""
        drainer = self._make_greenlet_drainer(app)
        drainer._shutdown.set()
        drainer._exc = None

        with pytest.raises(Exception, match=E_CELERY_RESTART_REQUIRED):
            drainer._ensure_not_shut_down()

    def test_ensure_not_shut_down_noop_when_running(self, app):
        """No-op when _shutdown is not set."""
        drainer = self._make_greenlet_drainer(app)
        # Should not raise
        drainer._ensure_not_shut_down()

    # -- start / stop -------------------------------------------------------

    def test_start_spawns_and_waits(self, app):
        """start() calls spawn(run) and waits for _started."""
        drainer = self._make_greenlet_drainer(app)

        def fake_spawn(func):
            # Run synchronously with immediate stop.
            drainer._stopped.set()
            func()

        drainer.spawn = fake_spawn
        drainer.result_consumer.drain_events = Mock(
            side_effect=lambda timeout=None: drainer._stopped.set()
        )
        drainer.start()

        assert drainer._started.is_set()
        assert drainer._shutdown.is_set()

    def test_start_raises_if_already_shut_down(self, app):
        """start() raises if drainer already completed."""
        drainer = self._make_greenlet_drainer(app)
        drainer._shutdown.set()

        with pytest.raises(Exception, match=E_CELERY_RESTART_REQUIRED):
            drainer.start()

    def test_stop_signals_and_waits(self, app):
        """stop() sets _stopped and waits for _shutdown."""
        drainer = self._make_greenlet_drainer(app)
        # Pre-set _shutdown so wait returns immediately.
        drainer._shutdown.set()
        drainer.stop()

        assert drainer._stopped.is_set()


# ---------------------------------------------------------------------------
# 3. Integration tests with real greenlet runtimes (gevent + eventlet)
# ---------------------------------------------------------------------------


@pytest.fixture(autouse=True)
def setup_eventlet():
    os.environ.update(EVENTLET_NO_GREENDNS='yes')


class DrainerTests:
    """
    Base test class for the Default / Gevent / Eventlet drainers.
    """

    interval = 0.1  # Check every tenth of a second
    MAX_TIMEOUT = 10  # Specify a max timeout so it doesn't run forever

    def get_drainer(self, environment):
        with patch('celery.backends.asynchronous.detect_environment') as d:
            d.return_value = environment
            backend = Backend(self.app)
            consumer = BaseResultConsumer(backend, self.app, backend.accept,
                                          pending_results={},
                                          pending_messages={})
            consumer.drain_events = Mock(side_effect=self.result_consumer_drain_events)
            return consumer.drainer

    @pytest.fixture(autouse=True)
    def setup_drainer(self):
        raise NotImplementedError

    @cached_property
    def sleep(self):
        """
        Sleep on the event loop.
        """
        raise NotImplementedError

    def schedule_thread(self, thread):
        """
        Set up a thread that runs on the event loop.
        """
        raise NotImplementedError

    def teardown_thread(self, thread):
        """
        Wait for a thread to stop.
        """
        raise NotImplementedError

    def result_consumer_drain_events(self, timeout=None):
        """
        Subclasses should override this method to define the behavior of
        drainer.result_consumer.drain_events.
        """
        raise NotImplementedError

    def test_drain_checks_on_interval(self):
        p = promise()

        def fulfill_promise_thread():
            self.sleep(self.interval * 2)
            p('done')

        fulfill_thread = self.schedule_thread(fulfill_promise_thread)

        on_interval = Mock()
        for _ in self.drainer.drain_events_until(p,
                                                 on_interval=on_interval,
                                                 interval=self.interval,
                                                 timeout=self.MAX_TIMEOUT):
            pass

        self.teardown_thread(fulfill_thread)

        assert p.ready, 'Should have terminated with promise being ready'
        assert on_interval.call_count < 20, 'Should have limited number of calls to on_interval'

    def test_drain_does_not_block_event_loop(self):
        """
        This test makes sure that other greenlets can still operate while drain_events_until is
        running.
        """
        p = promise()
        liveness_mock = Mock()

        def fulfill_promise_thread():
            self.sleep(self.interval * 2)
            p('done')

        def liveness_thread():
            while 1:
                if p.ready:
                    return
                self.sleep(self.interval / 10)
                liveness_mock()

        fulfill_thread = self.schedule_thread(fulfill_promise_thread)
        liveness_thread = self.schedule_thread(liveness_thread)

        on_interval = Mock()
        for _ in self.drainer.drain_events_until(p,
                                                 on_interval=on_interval,
                                                 interval=self.interval,
                                                 timeout=self.MAX_TIMEOUT):
            pass

        self.teardown_thread(fulfill_thread)
        self.teardown_thread(liveness_thread)

        assert p.ready, 'Should have terminated with promise being ready'
        assert on_interval.call_count <= liveness_mock.call_count, \
            'Should have served liveness_mock while waiting for event'

    def test_drain_timeout(self):
        p = promise()
        on_interval = Mock()

        with pytest.raises(socket.timeout):
            for _ in self.drainer.drain_events_until(p,
                                                     on_interval=on_interval,
                                                     interval=self.interval,
                                                     timeout=self.interval * 5):
                pass

        assert not p.ready, 'Promise should remain un-fulfilled'
        assert on_interval.call_count < 20, 'Should have limited number of calls to on_interval'

    def test_drain_catches_and_logs_oserror(self):
        p = promise()

        def fulfill():
            self.sleep(self.interval * 2)
            p('done')

        t = self.schedule_thread(fulfill)

        state = {'n': 0}

        def flaky(*args, **kwargs):
            state['n'] += 1
            if state['n'] == 1:
                raise OSError('simulated broker restart')
            # Yield to hub so the promise thread can run.
            self.result_consumer_drain_events(
                timeout=kwargs.get('timeout', None),
            )

        with patch.object(
            self.drainer.result_consumer, 'drain_events',
            side_effect=flaky,
        ):
            with patch('logging.warning') as mock_warn:
                for _ in self.drainer.drain_events_until(
                        p, interval=self.interval,
                        timeout=self.MAX_TIMEOUT):
                    pass

        self.teardown_thread(t)
        assert p.ready
        assert mock_warn.called


class GreenletDrainerTests(DrainerTests):
    def test_drain_raises_when_greenlet_already_exited(self):
        with patch.object(self.drainer.result_consumer, 'drain_events', side_effect=Exception("Test Exception")):
            thread = self.schedule_thread(self.drainer.run)

            with pytest.raises(Exception, match="Test Exception"):
                p = promise()

                for _ in self.drainer.drain_events_until(p, interval=self.interval):
                    pass

            self.teardown_thread(thread)

    def test_drain_raises_while_waiting_on_exiting_greenlet(self):
        with patch.object(self.drainer.result_consumer, 'drain_events', side_effect=Exception("Test Exception")):
            with pytest.raises(Exception, match="Test Exception"):
                p = promise()

                for _ in self.drainer.drain_events_until(p, interval=self.interval):
                    pass

    def test_start_raises_if_previous_error_in_run(self):
        with patch.object(self.drainer.result_consumer, 'drain_events', side_effect=Exception("Test Exception")):
            thread = self.schedule_thread(self.drainer.run)

            with pytest.raises(Exception, match="Test Exception"):
                self.drainer.start()

            self.teardown_thread(thread)

    def test_start_raises_if_drainer_already_stopped(self):
        with patch.object(self.drainer.result_consumer, 'drain_events', side_effect=lambda **_: self.sleep(0)):
            thread = self.schedule_thread(self.drainer.run)
            self.drainer.stop()

            with pytest.raises(Exception, match=E_CELERY_RESTART_REQUIRED):
                self.drainer.start()

            self.teardown_thread(thread)

    def test_run_catches_and_logs_oserror(self):
        def flaky(*args, **kwargs):
            if not hasattr(flaky, '_raised'):
                flaky._raised = True
                raise OSError('simulated broker restart in greenlet')
            self.drainer._stopped.set()

        with patch.object(
            self.drainer.result_consumer, 'drain_events',
            side_effect=flaky,
        ):
            with patch('logging.warning') as mock_warn:
                t = self.schedule_thread(self.drainer.run)
                self.teardown_thread(t)

        assert mock_warn.called
        assert 'connection error during drain_events' in mock_warn.call_args[0][0]
        assert self.drainer._exc is None


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="hangs forever intermittently on windows"
)
class test_EventletDrainer(GreenletDrainerTests):
    @pytest.fixture(autouse=True)
    def setup_drainer(self):
        pytest.importorskip('eventlet')
        self.drainer = self.get_drainer('eventlet')

    @cached_property
    def sleep(self):
        from eventlet import sleep
        return sleep

    def result_consumer_drain_events(self, timeout=None):
        import eventlet

        # `drain_events` of asynchronous backends with pubsub have to sleep
        # while waiting events for not more then `interval` timeout,
        # but events may coming sooner
        eventlet.sleep(timeout/10)

    def schedule_thread(self, thread):
        import eventlet
        g = eventlet.spawn(thread)
        eventlet.sleep(0)
        return g

    def teardown_thread(self, thread):
        try:
            # eventlet's wait() propagates any errors on the green thread, unlike
            # similar methods in gevent or python's threading library
            thread.wait()
        except Exception:
            pass


class test_Drainer(DrainerTests):
    @pytest.fixture(autouse=True)
    def setup_drainer(self):
        self.drainer = self.get_drainer('default')

    @cached_property
    def sleep(self):
        from time import sleep
        return sleep

    def result_consumer_drain_events(self, timeout=None):
        time.sleep(timeout)

    def schedule_thread(self, thread):
        t = threading.Thread(target=thread)
        t.start()
        return t

    def teardown_thread(self, thread):
        thread.join()


class test_GeventDrainer(GreenletDrainerTests):
    @pytest.fixture(autouse=True)
    def setup_drainer(self):
        pytest.importorskip('gevent')
        self.drainer = self.get_drainer('gevent')

    @cached_property
    def sleep(self):
        from gevent import sleep
        return sleep

    def result_consumer_drain_events(self, timeout=None):
        import gevent

        # `drain_events` of asynchronous backends with pubsub have to sleep
        # while waiting events for not more then `interval` timeout,
        # but events may coming sooner
        gevent.sleep(timeout/10)

    def schedule_thread(self, thread):
        import gevent
        g = gevent.spawn(thread)
        gevent.sleep(0)
        return g

    def teardown_thread(self, thread):
        import gevent
        gevent.wait([thread])


class test_BaseResultConsumer_reconnect:

    def _make_consumer(self, app):
        return _make_consumer(app)

    def test_reconnect_on_error_no_exception_passes_through(self, app):
        consumer = self._make_consumer(app)
        result = []
        with consumer.reconnect_on_error():
            result.append('ok')
        assert result == ['ok']

    def test_reconnect_on_error_ignores_non_connection_error(self, app):
        consumer = self._make_consumer(app)
        with pytest.raises(ValueError):
            with consumer.reconnect_on_error():
                raise ValueError('unrelated')

    def test_reconnect_on_error_default_connection_errors_empty(self, app):
        consumer = self._make_consumer(app)
        assert consumer._connection_errors == ()

        class FakeConnError(Exception):
            pass

        with pytest.raises(FakeConnError):
            with consumer.reconnect_on_error():
                raise FakeConnError('dropped')

    def test_reconnect_on_error_calls_reconnect_on_connection_error(self, app):
        consumer = self._make_consumer(app)

        class FakeConnError(Exception):
            pass

        consumer._connection_errors = (FakeConnError,)
        consumer._reconnect = Mock()

        with consumer.reconnect_on_error():
            raise FakeConnError('dropped')

        consumer._reconnect.assert_called_once_with()

    def test_reconnect_on_error_raises_runtime_when_reconnect_also_fails(self, app):
        consumer = self._make_consumer(app)

        class FakeConnError(Exception):
            pass

        consumer._connection_errors = (FakeConnError,)
        consumer._reconnect = Mock(side_effect=FakeConnError('still down'))

        with pytest.raises(RuntimeError, match='Retry limit exceeded'):
            with consumer.reconnect_on_error():
                raise FakeConnError('dropped')

    def test_reconnect_on_error_runtime_chained_from_connection_error(self, app):
        consumer = self._make_consumer(app)

        class FakeConnError(Exception):
            pass

        consumer._connection_errors = (FakeConnError,)
        original = FakeConnError('still down')
        consumer._reconnect = Mock(side_effect=original)

        with pytest.raises(RuntimeError) as exc_info:
            with consumer.reconnect_on_error():
                raise FakeConnError('dropped')

        assert exc_info.value.__cause__ is original

    def test_reconnect_base_implementation_is_noop(self, app):
        consumer = self._make_consumer(app)

        assert consumer._reconnect() is None



import itertools
import sys
from logging import LogRecord
from typing import Iterator
from unittest.mock import MagicMock, Mock, patch

if sys.version_info < (3, 11):  # pragma: no cover
    from exceptiongroup import ExceptionGroup

import pytest
from amqp import NotFound
from kombu import Exchange, Queue
from kombu.utils.functional import retry_over_time

from celery.worker.consumer.delayed_delivery import MAX_RETRIES, RETRY_INTERVAL, DelayedDelivery


class test_DelayedDelivery:
    @patch('celery.worker.consumer.delayed_delivery.detect_quorum_queues', return_value=[False, ""])
    def test_include_if_no_quorum_queues_detected(self, _):
        consumer_mock = Mock()

        delayed_delivery = DelayedDelivery(consumer_mock)

        assert delayed_delivery.include_if(consumer_mock) is False

    @patch('celery.worker.consumer.delayed_delivery.detect_quorum_queues', return_value=[True, ""])
    def test_include_if_quorum_queues_detected(self, _):
        consumer_mock = Mock()

        delayed_delivery = DelayedDelivery(consumer_mock)

        assert delayed_delivery.include_if(consumer_mock) is True

    def test_start_native_delayed_delivery_direct_exchange(self, caplog):
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='direct'))
        }

        delayed_delivery = DelayedDelivery(consumer_mock)

        delayed_delivery.start(consumer_mock)

        assert len(caplog.records) == 1
        record: LogRecord = caplog.records[0]
        assert record.levelname == "WARNING"
        assert record.message == (
            "Exchange celery is a direct exchange "
            "and native delayed delivery do not support direct exchanges.\n"
            "ETA tasks published to this exchange "
            "will block the worker until the ETA arrives."
        )

    def test_start_native_delayed_delivery_topic_exchange(self, caplog):
        consumer_mock = Mock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='topic'))
        }
        connection = MagicMock()
        consumer_mock.app.connection_for_write.return_value = connection

        delayed_delivery = DelayedDelivery(consumer_mock)

        delayed_delivery.start(consumer_mock)

        assert len(caplog.records) == 0
        # Verify connection context was called
        assert connection.__enter__.called
        assert connection.__exit__.called

    def test_start_native_delayed_delivery_fanout_exchange(self, caplog):
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='fanout'))
        }

        delayed_delivery = DelayedDelivery(consumer_mock)

        delayed_delivery.start(consumer_mock)

        assert len(caplog.records) == 0

    @pytest.mark.parametrize(
        "broker_urls, expected_result",
        [
            ("amqp://", {"amqp://"}),
            ("amqp://;redis://", {"amqp://", "redis://"}),
            (
                ["amqp://", "redis://", "sqs://"],
                {"amqp://", "redis://", "sqs://"},
            ),
        ],
    )
    def test_validate_broker_urls_valid(self, broker_urls, expected_result):
        delayed_delivery = DelayedDelivery(Mock())
        urls = delayed_delivery._validate_broker_urls(broker_urls)
        assert urls == expected_result

    @pytest.mark.parametrize(
        "broker_urls, exception_type, exception_match",
        [
            ("", ValueError, "broker_url configuration is empty"),
            (None, ValueError, "broker_url configuration is empty"),
            ([], ValueError, "broker_url configuration is empty"),
            (123, ValueError, "broker_url must be a string or list"),
            (["amqp://", 123, None, "amqp://"], ValueError, "All broker URLs must be strings"),
        ],
    )
    def test_validate_broker_urls_invalid(self, broker_urls, exception_type, exception_match):
        delayed_delivery = DelayedDelivery(Mock())
        with pytest.raises(exception_type, match=exception_match):
            delayed_delivery._validate_broker_urls(broker_urls)

    def test_validate_queue_type_empty(self):
        delayed_delivery = DelayedDelivery(Mock())

        with pytest.raises(ValueError, match="broker_native_delayed_delivery_queue_type is not configured"):
            delayed_delivery._validate_queue_type(None)

        with pytest.raises(ValueError, match="broker_native_delayed_delivery_queue_type is not configured"):
            delayed_delivery._validate_queue_type("")

    def test_validate_queue_type_invalid(self):
        delayed_delivery = DelayedDelivery(Mock())

        with pytest.raises(ValueError, match="Invalid queue type 'invalid'. Must be one of: classic, quorum"):
            delayed_delivery._validate_queue_type("invalid")

    def test_validate_queue_type_valid(self):
        delayed_delivery = DelayedDelivery(Mock())

        delayed_delivery._validate_queue_type("classic")
        delayed_delivery._validate_queue_type("quorum")

    @patch('celery.worker.consumer.delayed_delivery.retry_over_time')
    def test_start_retry_on_connection_error(self, mock_retry, caplog):
        consumer_mock = Mock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://localhost;amqp://backup'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='topic'))
        }

        mock_retry.side_effect = ConnectionRefusedError("Connection refused")

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        # Should try both URLs
        assert mock_retry.call_count == 2
        # Should log warning for each failed attempt
        assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 2
        # Should log critical when all URLs fail
        assert len([r for r in caplog.records if r.levelname == "CRITICAL"]) == 1

    def test_on_retry_logging(self, caplog):
        delayed_delivery = DelayedDelivery(Mock())
        exc = ConnectionRefusedError("Connection refused")

        # Create a dummy float iterator
        interval_range = iter([1.0, 2.0, 3.0])
        intervals_count = 1

        delayed_delivery._on_retry(exc, interval_range, intervals_count)

        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.levelname == "WARNING"
        assert "attempt 2/3" in record.message
        assert "Connection refused" in record.message

    def test_on_retry_argument_types(self):
        delayed_delivery_instance = DelayedDelivery(parent=Mock())
        fake_exception = ConnectionRefusedError("Simulated failure")

        # Define a custom errback to check types
        def type_checking_errback(self, exc, interval_range, intervals_count):
            assert isinstance(exc, Exception), f"Expected Exception, got {type(exc)}"
            assert isinstance(interval_range, Iterator), f"Expected Iterator, got {type(interval_range)}"
            assert isinstance(intervals_count, int), f"Expected int, got {type(intervals_count)}"

            peek_iter, interval_range = itertools.tee(interval_range)
            try:
                first = next(peek_iter)
                assert isinstance(first, float)
            except StopIteration:
                pass

            return 0.1

        # Patch _setup_delayed_delivery to raise the exception immediately
        with patch.object(delayed_delivery_instance, '_setup_delayed_delivery', side_effect=fake_exception):
            # Patch _on_retry properly as a bound method to avoid 'missing self'
            with patch.object(
                delayed_delivery_instance,
                '_on_retry',
                new=type_checking_errback.__get__(delayed_delivery_instance)
            ):
                try:
                    with pytest.raises(ConnectionRefusedError):
                        retry_over_time(
                            delayed_delivery_instance._setup_delayed_delivery,
                            args=(Mock(), "amqp://localhost"),
                            catch=(ConnectionRefusedError,),
                            errback=delayed_delivery_instance._on_retry,
                            interval_start=RETRY_INTERVAL,
                            max_retries=MAX_RETRIES,
                        )
                except ConnectionRefusedError:
                    pass  # expected

    def test_retry_over_time_with_float_return(self):
        delayed_delivery = DelayedDelivery(parent=Mock())
        return_values = []

        # Wrap the real _on_retry method to capture its return value
        original_on_retry = delayed_delivery._on_retry

        def wrapped_on_retry(exc, interval_range, intervals_count):
            result = original_on_retry(exc, interval_range, intervals_count)
            return_values.append(result)
            return result

        with patch.object(
            delayed_delivery, '_setup_delayed_delivery',
            side_effect=ConnectionRefusedError("Simulated failure")
        ):
            with pytest.raises(ConnectionRefusedError):
                retry_over_time(
                    fun=delayed_delivery._setup_delayed_delivery,
                    args=(Mock(), "amqp://localhost"),
                    catch=(ConnectionRefusedError,),
                    errback=wrapped_on_retry,
                    interval_start=RETRY_INTERVAL,
                    max_retries=MAX_RETRIES
                )

        assert len(return_values) == MAX_RETRIES
        for value in return_values:
            assert isinstance(value, float), f"Expected float, got {type(value)}"

    def test_start_with_no_queues(self, caplog):
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {}

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 1
        assert "No queues found to bind for delayed delivery" in caplog.records[0].message

    def test_start_configuration_validation_error(self, caplog):
        consumer_mock = Mock()
        consumer_mock.app.conf.broker_url = ""  # Invalid broker URL

        delayed_delivery = DelayedDelivery(consumer_mock)

        with pytest.raises(ValueError, match="broker_url configuration is empty"):
            delayed_delivery.start(consumer_mock)

        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.levelname == "CRITICAL"
        assert "Configuration validation failed" in record.message

    @patch('celery.worker.consumer.delayed_delivery.declare_native_delayed_delivery_exchanges_and_queues')
    def test_setup_declare_error(self, mock_declare, caplog):
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='topic'))
        }

        mock_declare.side_effect = Exception("Failed to declare")

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        # Should log warning and critical messages
        assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 2
        assert len([r for r in caplog.records if r.levelname == "CRITICAL"]) == 1
        assert any("Failed to declare exchanges and queues" in r.message for r in caplog.records)
        assert any("Failed to setup delayed delivery for all broker URLs" in r.message for r in caplog.records)

    @patch('celery.worker.consumer.delayed_delivery.bind_queue_to_native_delayed_delivery_exchange')
    def test_setup_bind_error(self, mock_bind, caplog):
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = 'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'
        consumer_mock.app.amqp.queues = {
            'celery': Queue('celery', exchange=Exchange('celery', type='topic'))
        }

        mock_bind.side_effect = Exception("Failed to bind")

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        # Should log warning and critical messages
        assert len([r for r in caplog.records if r.levelname == "WARNING"]) == 2
        assert len([r for r in caplog.records if r.levelname == "CRITICAL"]) == 1
        assert any("Failed to bind queue" in r.message for r in caplog.records)
        assert any("Failed to setup delayed delivery for all broker URLs" in r.message for r in caplog.records)

    @patch('celery.worker.consumer.delayed_delivery.bind_queue_to_native_delayed_delivery_exchange')
    def test_bind_queues_continues_after_failure(self, mock_bind, caplog):
        """
        Test that binding continues for remaining queues after one fails.
        """
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = \
            'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'

        # Create three queues
        queue1 = Queue('queue1', exchange=Exchange('exchange1', type='topic'))
        queue2 = Queue('queue2', exchange=Exchange('exchange2', type='topic'))
        queue3 = Queue('queue3', exchange=Exchange('exchange3', type='topic'))

        consumer_mock.app.amqp.queues = {
            'queue1': queue1,
            'queue2': queue2,
            'queue3': queue3,
        }

        # Make the second queue fail to bind
        def bind_side_effect(connection, queue):
            if queue.name == 'queue2':
                raise NotFound(
                    reply_text="NOT_FOUND - no queue 'queue2' in vhost '/'",
                    method_name="Queue.bind",
                    reply_code=404,
                )

        mock_bind.side_effect = bind_side_effect

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        # Verify that bind was called for all three queues
        assert mock_bind.call_count == 3

        # Verify error was logged for queue2
        error_logs = [r for r in caplog.records if r.levelname == "ERROR"]
        expected_msg = \
            "Queue.bind: (404) NOT_FOUND - no queue 'queue2' in vhost '/'"
        assert any(expected_msg in r.message for r in error_logs)

    @patch('celery.worker.consumer.delayed_delivery.bind_queue_to_native_delayed_delivery_exchange')
    def test_bind_queues_raises_exceptions_on_failures(self, mock_bind):
        """Test that Exceptions are raised with all binding failures."""
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = \
            'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'

        # Create three queues
        queue1 = Queue('queue1', exchange=Exchange('exchange1', type='topic'))
        queue2 = Queue('queue2', exchange=Exchange('exchange2', type='topic'))
        queue3 = Queue('queue3', exchange=Exchange('exchange3', type='topic'))

        consumer_mock.app.amqp.queues = {
            'queue1': queue1,
            'queue2': queue2,
            'queue3': queue3,
        }

        # Make queue1 and queue3 fail with different errors
        def bind_side_effect(connection, queue):
            if queue.name == 'queue1':
                raise ValueError("Queue1 binding failed")
            elif queue.name == 'queue3':
                raise RuntimeError("Queue3 binding failed")

        mock_bind.side_effect = bind_side_effect

        delayed_delivery = DelayedDelivery(consumer_mock)

        # Should raise RuntimeError containing both exceptions messages
        with pytest.raises(ExceptionGroup) as exc_info:
            delayed_delivery._setup_delayed_delivery(consumer_mock, 'amqp://')

        # Verify the RuntimeError message contains both exceptions
        exception_group = exc_info.value
        assert str(exception_group) == (
            "One or more failures occurred while binding queues to delayed "
            "delivery exchanges (2 sub-exceptions)"
        )

        # Verify the ExceptionGroup contains both exceptions
        exceptions = exception_group.exceptions
        assert len(exceptions) == 2
        assert any(
            isinstance(ex, ValueError) and str(ex) == "Queue1 binding failed"
            for ex in exceptions
        )
        assert any(
            isinstance(ex, RuntimeError) and str(ex) == "Queue3 binding failed"
            for ex in exceptions
        )

        # Verify bind was called for all three queues
        assert mock_bind.call_count == 3

    @patch('celery.worker.consumer.delayed_delivery.bind_queue_to_native_delayed_delivery_exchange')
    def test_bind_retries_on_retried_exception(self, mock_bind, caplog):
        """
        Test that retried exceptions from
        bind_queue_to_native_delayed_delivery_exchange trigger the retry
        mechanism.
        """
        consumer_mock = MagicMock()
        consumer_mock.app.conf.broker_native_delayed_delivery_queue_type = \
            'classic'
        consumer_mock.app.conf.broker_url = 'amqp://'

        # Create a queue
        queue1 = Queue('queue1', exchange=Exchange('exchange1', type='topic'))
        consumer_mock.app.amqp.queues = {'queue1': queue1}

        # Track bind attempts
        bind_attempts = [0]

        # Make bind raise a ConnectionRefusedError twice, then succeed
        # This simulates a transient connection issue that resolves on retry
        def bind_side_effect(connection, queue):
            bind_attempts[0] += 1
            if bind_attempts[0] <= 2:
                # ConnectionRefusedError is one of the RETRIED_EXCEPTIONS
                raise ConnectionRefusedError("Connection refused")
            # Succeed on third attempt

        mock_bind.side_effect = bind_side_effect

        delayed_delivery = DelayedDelivery(consumer_mock)
        delayed_delivery.start(consumer_mock)

        # Verify bind was attempted multiple times (indicating retries
        # occurred)
        assert bind_attempts[0] == 3, \
            "Expected 3 bind attempts (2 failures + 1 success), got " + \
            f"{bind_attempts[0]}"

        # Verify retry warnings were logged
        warning_logs = [r for r in caplog.records if r.levelname == "WARNING"]
        retry_warnings = [
            r for r in warning_logs
            if "Retrying delayed delivery setup" in r.message
        ]

        # Should have 2 retry warnings (one for each failed attempt)
        assert len(retry_warnings) == 2, \
            f"Expected 2 retry warnings, got {len(retry_warnings)}. " + \
            f"All warnings: {[r.message for r in warning_logs]}"

        # Verify the retry messages contain the expected information and
        # correct attempt numbers
        assert "Connection refused" in retry_warnings[0].message
        assert "attempt 1/" in retry_warnings[0].message
        assert "Connection refused" in retry_warnings[1].message
        assert "attempt 2/" in retry_warnings[1].message



from datetime import datetime
from pickle import dumps, loads
from unittest.mock import Mock

import pytest

from celery import states
from celery.exceptions import ImproperlyConfigured
from celery.utils.objects import Bunch

CASSANDRA_MODULES = [
    'cassandra',
    'cassandra.auth',
    'cassandra.cluster',
    'cassandra.query',
]


class test_CassandraBackend:

    def setup_method(self):
        self.app.conf.update(
            cassandra_servers=['example.com'],
            cassandra_keyspace='celery',
            cassandra_table='task_results',
        )

    @pytest.mark.patched_module(*CASSANDRA_MODULES)
    def test_init_no_cassandra(self, module):
        # should raise ImproperlyConfigured when no python-driver
        # installed.
        from celery.backends import cassandra as mod
        prev, mod.cassandra = mod.cassandra, None
        try:
            with pytest.raises(ImproperlyConfigured):
                mod.CassandraBackend(app=self.app)
        finally:
            mod.cassandra = prev

    @pytest.mark.patched_module(*CASSANDRA_MODULES)
    def test_init_with_and_without_LOCAL_QUROM(self, module):
        from celery.backends import cassandra as mod
        mod.cassandra = Mock()

        cons = mod.cassandra.ConsistencyLevel = Bunch(
            LOCAL_QUORUM='foo',
        )

        self.app.conf.cassandra_read_consistency = 'LOCAL_FOO'
        self.app.conf.cassandra_write_consistency = 'LOCAL_FOO'

        mod.CassandraBackend(app=self.app)
        cons.LOCAL_FOO = 'bar'
        mod.CassandraBackend(app=self.app)

        # no servers and no bundle_path raises ImproperlyConfigured
        with pytest.raises(ImproperlyConfigured):
            self.app.conf.cassandra_servers = None
            self.app.conf.cassandra_secure_bundle_path = None
            mod.CassandraBackend(
                app=self.app, keyspace='b', column_family='c',
            )

        # both servers no bundle_path raises ImproperlyConfigured
        with pytest.raises(ImproperlyConfigured):
            self.app.conf.cassandra_servers = ['localhost']
            self.app.conf.cassandra_secure_bundle_path = (
                '/home/user/secure-connect-bundle.zip')
            mod.CassandraBackend(
                app=self.app, keyspace='b', column_family='c',
            )

    def test_init_with_cloud(self):
        # Tests behavior when Cluster.connect works properly
        # and cluster is created with 'cloud' param instead of 'contact_points'
        from celery.backends import cassandra as mod

        class DummyClusterWithBundle:

            def __init__(self, *args, **kwargs):
                if args != ():
                    # this cluster is supposed to be created with 'cloud=...'
                    raise ValueError('I should be created with kwargs only')
                pass

            def connect(self, *args, **kwargs):
                return Mock()

        mod.cassandra = Mock()
        mod.cassandra.cluster = Mock()
        mod.cassandra.cluster.Cluster = DummyClusterWithBundle

        self.app.conf.cassandra_secure_bundle_path = '/path/to/bundle.zip'
        self.app.conf.cassandra_servers = None

        x = mod.CassandraBackend(app=self.app)
        x._get_connection()
        assert isinstance(x._cluster, DummyClusterWithBundle)

    @pytest.mark.patched_module(*CASSANDRA_MODULES)
    @pytest.mark.usefixtures('depends_on_current_app')
    def test_reduce(self, module):
        from celery.backends.cassandra import CassandraBackend
        assert loads(dumps(CassandraBackend(app=self.app)))

    @pytest.mark.patched_module(*CASSANDRA_MODULES)
    def test_get_task_meta_for(self, module):
        from celery.backends import cassandra as mod
        mod.cassandra = Mock()

        x = mod.CassandraBackend(app=self.app)
        session = x._session = Mock()
        execute = session.execute = Mock()
        result_set = Mock()
        result_set.one.return_value = [
            states.SUCCESS, '1', datetime.now(), b'', b''
        ]
        execute.return_value = result_set
        x.decode = Mock()
        meta = x._get_task_meta_for('task_id')
        assert meta['status'] == states.SUCCESS

        result_set.one.return_value = []
        x._session.execute.return_value = result_set
        meta = x._get_task_meta_for('task_id')
        assert meta['status'] == states.PENDING

    def test_as_uri(self):
        # Just ensure as_uri works properly
        from celery.backends import cassandra as mod
        mod.cassandra = Mock()

        x = mod.CassandraBackend(app=self.app)
        x.as_uri()
        x.as_uri(include_password=False)

    @pytest.mark.patched_module(*CASSANDRA_MODULES)
    def test_store_result(self, module):
        from celery.backends import cassandra as mod
        mod.cassandra = Mock()

        x = mod.CassandraBackend(app=self.app)
        session = x._session = Mock()
        session.execute = Mock()
        x._store_result('task_id', 'result', states.SUCCESS)

    def test_timeouting_cluster(self):
        # Tests behavior when Cluster.connect raises
        # cassandra.OperationTimedOut.
        from celery.backends import cassandra as mod

        class OTOExc(Exception):
            pass

        class VeryFaultyCluster:
            def __init__(self, *args, **kwargs):
                pass

            def connect(self, *args, **kwargs):
                raise OTOExc()

            def shutdown(self):
                pass

        mod.cassandra = Mock()
        mod.cassandra.OperationTimedOut = OTOExc
        mod.cassandra.cluster = Mock()
        mod.cassandra.cluster.Cluster = VeryFaultyCluster

        x = mod.CassandraBackend(app=self.app)

        with pytest.raises(OTOExc):
            x._store_result('task_id', 'result', states.SUCCESS)
        assert x._cluster is None
        assert x._session is None

    def test_create_result_table(self):
        # Tests behavior when session.execute raises
        # cassandra.AlreadyExists.
        from celery.backends import cassandra as mod

        class OTOExc(Exception):
            pass

        class FaultySession:
            def __init__(self, *args, **kwargs):
                pass

            def execute(self, *args, **kwargs):
                raise OTOExc()

        class DummyCluster:

            def __init__(self, *args, **kwargs):
                pass

            def connect(self, *args, **kwargs):
                return FaultySession()

        mod.cassandra = Mock()
        mod.cassandra.cluster = Mock()
        mod.cassandra.cluster.Cluster = DummyCluster
        mod.cassandra.AlreadyExists = OTOExc

        x = mod.CassandraBackend(app=self.app)
        x._get_connection(write=True)
        assert x._session is not None

    def test_init_session(self):
        # Tests behavior when Cluster.connect works properly
        from celery.backends import cassandra as mod

        class DummyCluster:

            def __init__(self, *args, **kwargs):
                pass

            def connect(self, *args, **kwargs):
                return Mock()

        mod.cassandra = Mock()
        mod.cassandra.cluster = Mock()
        mod.cassandra.cluster.Cluster = DummyCluster

        x = mod.CassandraBackend(app=self.app)
        assert x._session is None
        x._get_connection(write=True)
        assert x._session is not None

        s = x._session
        x._get_connection()
        assert s is x._session

    def test_auth_provider(self):
        # Ensure valid auth_provider works properly, and invalid one raises
        # ImproperlyConfigured exception.
        from celery.backends import cassandra as mod

        class DummyAuth:
            ValidAuthProvider = Mock()

        mod.cassandra = Mock()
        mod.cassandra.auth = DummyAuth

        # Valid auth_provider
        self.app.conf.cassandra_auth_provider = 'ValidAuthProvider'
        self.app.conf.cassandra_auth_kwargs = {
            'username': 'stuff'
        }
        mod.CassandraBackend(app=self.app)

        # Invalid auth_provider
        self.app.conf.cassandra_auth_provider = 'SpiderManAuth'
        self.app.conf.cassandra_auth_kwargs = {
            'username': 'Jack'
        }
        with pytest.raises(ImproperlyConfigured):
            mod.CassandraBackend(app=self.app)

    def test_options(self):
        # Ensure valid options works properly
        from celery.backends import cassandra as mod

        mod.cassandra = Mock()
        # Valid options
        self.app.conf.cassandra_options = {
            'cql_version': '3.2.1',
            'protocol_version': 3
        }
        self.app.conf.cassandra_port = None
        x = mod.CassandraBackend(app=self.app)
        # Default port is 9042
        assert x.port == 9042

        # Valid options with port specified
        self.app.conf.cassandra_port = 1234
        x = mod.CassandraBackend(app=self.app)
        assert x.port == 1234



import gc
import logging
import os
import pprint
import uuid

import billiard as multiprocessing
import pytest
from kombu import Queue
from kombu.pools import connections

from celery import Celery, _state
from celery.contrib.testing.worker import start_worker

QOS_GLOBAL_ERROR = "qos.global not allowed"

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_app(queue_name: str) -> Celery:
    rabbitmq_user = os.environ.get("RABBITMQ_DEFAULT_USER", "guest")
    rabbitmq_pass = os.environ.get("RABBITMQ_DEFAULT_PASS", "guest")
    redis_host = os.environ.get("REDIS_HOST", "localhost")
    redis_port = os.environ.get("REDIS_PORT", "6379")

    broker_url = os.environ.get("TEST_BROKER", f"pyamqp://{rabbitmq_user}:{rabbitmq_pass}@localhost:5672//")
    backend_url = os.environ.get("TEST_BACKEND", f"redis://{redis_host}:{redis_port}/0")

    app = Celery("quorum_qos_race", broker=broker_url, backend=backend_url)

    app.conf.task_queues = [
        Queue(
            name=queue_name,
            queue_arguments={"x-queue-type": "quorum"},
        )
    ]
    app.conf.task_default_queue = queue_name
    app.conf.worker_prefetch_multiplier = 1
    app.conf.task_acks_late = True
    app.conf.task_reject_on_worker_lost = True
    app.conf.broker_transport_options = {"confirm_publish": True}

    return app


def dummy_task_factory(app: Celery, simulate_qos_issue: bool):
    @app.task(name="dummy_task")
    def dummy_task():
        if simulate_qos_issue:
            raise Exception("qos.global not allowed on quorum queues (simulated)")
        return "ok"
    return dummy_task


def run_worker(simulate_qos_issue: bool, result_queue: multiprocessing.Queue):
    queue_name = f"race_quorum_queue_{uuid.uuid4().hex}"
    app = create_app(queue_name)
    logger.info("[Celery config snapshot]:\n%s", pprint.pformat(dict(app.conf)))
    task = dummy_task_factory(app, simulate_qos_issue)

    try:
        with start_worker(
            app,
            queues=[queue_name],
            loglevel="INFO",
            perform_ping_check=False,
            shutdown_timeout=15,
        ):
            res = task.delay()
            try:
                result = res.get(timeout=10)
                result_queue.put({"status": "ok", "result": result})
            except Exception as e:
                result_queue.put({"status": "error", "reason": str(e)})
    except Exception as e:
        logger.exception("[worker %s] external failure", simulate_qos_issue)
        result_queue.put({"status": "external_failure", "reason": str(e)})
    finally:
        if result_queue.empty():
            result_queue.put({"status": "crash", "reason": "Worker crashed without reporting"})


@pytest.mark.amqp
@pytest.mark.timeout(90)
def test_rabbitmq_quorum_qos_visibility_race():
    try:
        multiprocessing.set_start_method("spawn", force=True)
    except RuntimeError:
        pass

    results = []
    processes = []
    queues = []

    for i in range(3):
        simulate = (i == 0)
        q = multiprocessing.Queue()
        queues.append(q)

        p = multiprocessing.Process(target=run_worker, args=(simulate, q))
        p.daemon = True
        processes.append(p)
        p.start()

    try:
        for i, (p, q) in enumerate(zip(processes, queues)):
            try:
                p.join(timeout=30)
                if p.is_alive():
                    p.terminate()
                    p.join(timeout=10)
                    results.append({"status": "timeout", "reason": f"[worker {i}] timeout"})
                else:
                    try:
                        results.append(q.get(timeout=5))
                    except Exception as e:
                        results.append({"status": "error", "reason": f"Result error: {str(e)}"})
            except Exception:
                try:
                    results.append(q.get(timeout=5))
                except Exception:
                    results.append({"status": "crash", "reason": f"Worker {i} crashed and gave no result"})

        if any(QOS_GLOBAL_ERROR in r.get("reason", "").lower() for r in results):
            pytest.xfail("Detected global QoS usage on quorum queue (simulated failure)")
    finally:
        for i, p in enumerate(processes):
            if p.is_alive():
                p.terminate()
                p.join(timeout=10)

        # Reset Kombu connection pools (safe public API)
        try:
            connections.clear()
        except Exception:
            pass

        # Reset Celery app/task global state
        _state._set_current_app(None)
        _state._task_stack.__init__()  # reinitialize stack to avoid stale state

        # Force garbage collection
        gc.collect()

        # Reset multiprocessing to default (may help restore test_multiprocess_producer expectations)
        if multiprocessing.get_start_method(allow_none=True) == "spawn":
            try:
                multiprocessing.set_start_method("fork", force=True)
            except RuntimeError:
                pass



import logging
import sys
from collections import defaultdict
from io import StringIO
from tempfile import mkstemp
from unittest.mock import Mock, patch

import pytest

from celery import signals, uuid
from celery.app.log import TaskFormatter
from celery.utils.log import ColorFormatter, LoggingProxy, get_logger, get_task_logger, in_sighandler
from celery.utils.log import logger as base_logger
from celery.utils.log import logger_isa, task_logger
from t.unit import conftest


class test_TaskFormatter:

    def test_no_task(self):
        class Record:
            msg = 'hello world'
            levelname = 'info'
            exc_text = exc_info = None
            stack_info = None

            def getMessage(self):
                return self.msg
        record = Record()
        x = TaskFormatter()
        x.format(record)
        assert record.task_name == '???'
        assert record.task_id == '???'


class test_logger_isa:

    def test_isa(self):
        x = get_task_logger('Z1george')
        assert logger_isa(x, task_logger)
        prev_x, x.parent = x.parent, None
        try:
            assert not logger_isa(x, task_logger)
        finally:
            x.parent = prev_x

        y = get_task_logger('Z1elaine')
        y.parent = x
        assert logger_isa(y, task_logger)
        assert logger_isa(y, x)
        assert logger_isa(y, y)

        z = get_task_logger('Z1jerry')
        z.parent = y
        assert logger_isa(z, task_logger)
        assert logger_isa(z, y)
        assert logger_isa(z, x)
        assert logger_isa(z, z)

    def test_recursive(self):
        x = get_task_logger('X1foo')
        prev, x.parent = x.parent, x
        try:
            with pytest.raises(RuntimeError):
                logger_isa(x, task_logger)
        finally:
            x.parent = prev

        y = get_task_logger('X2foo')
        z = get_task_logger('X2foo')
        prev_y, y.parent = y.parent, z
        try:
            prev_z, z.parent = z.parent, y
            try:
                with pytest.raises(RuntimeError):
                    logger_isa(y, task_logger)
            finally:
                z.parent = prev_z
        finally:
            y.parent = prev_y


class test_ColorFormatter:

    @patch('celery.utils.log.safe_str')
    @patch('logging.Formatter.formatException')
    def test_formatException_not_string(self, fe, safe_str):
        x = ColorFormatter()
        value = KeyError()
        fe.return_value = value
        assert x.formatException(value) is value
        fe.assert_called()
        safe_str.assert_not_called()

    @patch('logging.Formatter.formatException')
    @patch('celery.utils.log.safe_str')
    def test_formatException_bytes(self, safe_str, fe):
        x = ColorFormatter()
        fe.return_value = b'HELLO'
        try:
            raise Exception()
        except Exception:
            assert x.formatException(sys.exc_info())

    @patch('logging.Formatter.format')
    def test_format_object(self, _format):
        x = ColorFormatter()
        x.use_color = True
        record = Mock()
        record.levelname = 'ERROR'
        record.msg = object()
        assert x.format(record)

    @patch('celery.utils.log.safe_str')
    def test_format_raises(self, safe_str):
        x = ColorFormatter()

        def on_safe_str(s):
            try:
                raise ValueError('foo')
            finally:
                safe_str.side_effect = None
        safe_str.side_effect = on_safe_str

        class Record:
            levelname = 'ERROR'
            msg = 'HELLO'
            exc_info = 1
            exc_text = 'error text'
            stack_info = None

            def __str__(self):
                return on_safe_str('')

            def getMessage(self):
                return self.msg

        record = Record()
        safe_str.return_value = record

        msg = x.format(record)
        assert '<Unrepresentable' in msg
        assert safe_str.call_count == 1


class test_default_logger:

    def setup_logger(self, *args, **kwargs):
        self.app.log.setup_logging_subsystem(*args, **kwargs)

        return logging.root

    def setup_method(self):
        self.get_logger = lambda n=None: get_logger(n) if n else logging.root
        signals.setup_logging.receivers[:] = []
        self.app.log.already_setup = False

    def test_get_logger_sets_parent(self):
        logger = get_logger('celery.test_get_logger')
        assert logger.parent.name == base_logger.name

    def test_get_logger_root(self):
        logger = get_logger(base_logger.name)
        assert logger.parent is logging.root

    def test_setup_logging_subsystem_misc(self, restore_logging):
        self.app.log.setup_logging_subsystem(loglevel=None)

    def test_setup_logging_subsystem_misc2(self, restore_logging):
        self.app.conf.worker_hijack_root_logger = True
        self.app.log.setup_logging_subsystem()

    def test_get_default_logger(self):
        assert self.app.log.get_default_logger()

    def test_configure_logger(self):
        logger = self.app.log.get_default_logger()
        self.app.log._configure_logger(logger, sys.stderr, None, '', False)
        self.app.log._configure_logger(None, sys.stderr, None, '', False)
        logger.handlers[:] = []

    def test_setup_logging_subsystem_colorize(self, restore_logging):
        self.app.log.setup_logging_subsystem(colorize=None)
        self.app.log.setup_logging_subsystem(colorize=True)

    @pytest.mark.masked_modules('billiard.util')
    def test_setup_logging_subsystem_no_mputil(self, restore_logging, mask_modules):
        self.app.log.setup_logging_subsystem()

    def test_setup_logger(self, restore_logging):
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False, colorize=True)
        logger.handlers = []
        self.app.log.already_setup = False
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False, colorize=None)
        # setup_logger logs to stderr without logfile argument.
        assert (conftest.get_logger_handlers(logger)[0].stream is
                sys.__stderr__)

    def test_setup_logger_no_handlers_stream(self, restore_logging):
        l = self.get_logger()
        l.handlers = []

        with conftest.stdouts() as (stdout, stderr):
            l = self.setup_logger(logfile=sys.stderr,
                                  loglevel=logging.INFO, root=False)
            l.info('The quick brown fox...')
            assert 'The quick brown fox...' in stderr.getvalue()

    @patch('os.fstat')
    def test_setup_logger_no_handlers_file(self, *args):
        _, tempfile = mkstemp(suffix='unittest', prefix='celery')
        with patch('builtins.open') as osopen:
            with conftest.restore_logging_context_manager():
                files = defaultdict(StringIO)

                def open_file(filename, *args, **kwargs):
                    f = files[filename]
                    f.fileno = Mock()
                    f.fileno.return_value = 99
                    return f

                osopen.side_effect = open_file
                l = self.get_logger()
                l.handlers = []
                l = self.setup_logger(
                    logfile=tempfile, loglevel=logging.INFO, root=False,
                )
                assert isinstance(conftest.get_logger_handlers(l)[0],
                                  logging.FileHandler)
                assert tempfile in files

    def test_redirect_stdouts(self, restore_logging):
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False)
        try:
            with conftest.wrap_logger(logger) as sio:
                self.app.log.redirect_stdouts_to_logger(
                    logger, loglevel=logging.ERROR,
                )
                logger.error('foo')
                assert 'foo' in sio.getvalue()
                self.app.log.redirect_stdouts_to_logger(
                    logger, stdout=False, stderr=False,
                )
        finally:
            sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__

    def test_logging_proxy(self, restore_logging):
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False)

        with conftest.wrap_logger(logger) as sio:
            p = LoggingProxy(logger, loglevel=logging.ERROR)
            p.close()
            p.write('foo')
            assert 'foo' not in sio.getvalue()
            p.closed = False
            p.write('\n')
            assert sio.getvalue() == ''
            write_res = p.write('foo ')
            assert sio.getvalue() == 'foo \n'
            assert write_res == 4
            lines = ['baz', 'xuzzy']
            p.writelines(lines)
            for line in lines:
                assert line in sio.getvalue()
            p.flush()
            p.close()
            assert not p.isatty()

            with conftest.stdouts() as (stdout, stderr):
                with in_sighandler():
                    p.write('foo')
                    assert stderr.getvalue()

    def test_logging_proxy_bytes(self, restore_logging):
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False)

        with conftest.wrap_logger(logger) as sio:
            p = LoggingProxy(logger, loglevel=logging.ERROR)
            p.close()
            p.write(b'foo')
            assert 'foo' not in str(sio.getvalue())
            p.closed = False
            p.write(b'\n')
            assert str(sio.getvalue()) == ''
            write_res = p.write(b'foo ')
            assert str(sio.getvalue()) == 'foo \n'
            assert write_res == 4
            p.flush()
            p.close()
            assert not p.isatty()

            with conftest.stdouts() as (stdout, stderr):
                with in_sighandler():
                    p.write(b'foo')
                    assert stderr.getvalue()

    def test_logging_proxy_recurse_protection(self, restore_logging):
        logger = self.setup_logger(loglevel=logging.ERROR, logfile=None,
                                   root=False)
        p = LoggingProxy(logger, loglevel=logging.ERROR)
        p._thread.recurse_protection = True
        try:
            assert p.write('FOOFO') == 0
        finally:
            p._thread.recurse_protection = False


class test_task_logger(test_default_logger):

    def setup_method(self):
        logger = self.logger = get_logger('celery.task')
        logger.handlers = []
        logging.root.manager.loggerDict.pop(logger.name, None)
        self.uid = uuid()

        @self.app.task(shared=False)
        def test_task():
            pass
        self.get_logger().handlers = []
        self.task = test_task
        from celery._state import _task_stack
        _task_stack.push(test_task)

    def teardown_method(self):
        from celery._state import _task_stack
        _task_stack.pop()

    def setup_logger(self, *args, **kwargs):
        return self.app.log.setup_task_loggers(*args, **kwargs)

    def get_logger(self, *args, **kwargs):
        return get_task_logger('test_task_logger')

    def test_renaming_base_logger(self):
        with pytest.raises(RuntimeError):
            get_task_logger('celery')

    def test_renaming_task_logger(self):
        with pytest.raises(RuntimeError):
            get_task_logger('celery.task')


class MockLogger(logging.Logger):
    _records = None

    def __init__(self, *args, **kwargs):
        self._records = []
        super().__init__(*args, **kwargs)

    def handle(self, record):
        self._records.append(record)

    def isEnabledFor(self, level):
        return True



import collections
import sys

import pytest
from kombu.utils.functional import lazy

from celery.utils.functional import (DummyContext, first, firstmethod, fun_accepts_kwargs, fun_takes_argument,
                                     head_from_fun, is_numeric_value, lookahead, maybe_list, mlazy, padlist, regen,
                                     seq_concat_item, seq_concat_seq)


def test_DummyContext():
    with DummyContext():
        pass
    with pytest.raises(KeyError):
        with DummyContext():
            raise KeyError()


@pytest.mark.parametrize('items,n,default,expected', [
    (['George', 'Costanza', 'NYC'], 3, None,
     ['George', 'Costanza', 'NYC']),
    (['George', 'Costanza'], 3, None,
     ['George', 'Costanza', None]),
    (['George', 'Costanza', 'NYC'], 4, 'Earth',
     ['George', 'Costanza', 'NYC', 'Earth']),
])
def test_padlist(items, n, default, expected):
    assert padlist(items, n, default=default) == expected


class test_firstmethod:

    def test_AttributeError(self):
        assert firstmethod('foo')([object()]) is None

    def test_handles_lazy(self):

        class A:

            def __init__(self, value=None):
                self.value = value

            def m(self):
                return self.value

        assert 'four' == firstmethod('m')([
            A(), A(), A(), A('four'), A('five')])
        assert 'four' == firstmethod('m')([
            A(), A(), A(), lazy(lambda: A('four')), A('five')])


def test_first():
    iterations = [0]

    def predicate(value):
        iterations[0] += 1
        if value == 5:
            return True
        return False

    assert first(predicate, range(10)) == 5
    assert iterations[0] == 6

    iterations[0] = 0
    assert first(predicate, range(10, 20)) is None
    assert iterations[0] == 10


def test_lookahead():
    assert list(lookahead(x for x in range(6))) == [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, None)]


def test_maybe_list():
    assert maybe_list(1) == [1]
    assert maybe_list([1]) == [1]
    assert maybe_list(None) is None


def test_mlazy():
    it = iter(range(20, 30))
    p = mlazy(it.__next__)
    assert p() == 20
    assert p.evaluated
    assert p() == 20
    assert repr(p) == '20'


class test_regen:

    def test_list(self):
        l = [1, 2]
        r = regen(iter(l))
        assert regen(l) is l
        assert r == l
        assert r == l  # again
        assert r.__length_hint__() == 0

        fun, args = r.__reduce__()
        assert fun(*args) == l

    @pytest.fixture
    def g(self):
        return regen(iter(list(range(10))))

    def test_gen(self, g):
        assert g[7] == 7
        assert g[6] == 6
        assert g[5] == 5
        assert g[4] == 4
        assert g[3] == 3
        assert g[2] == 2
        assert g[1] == 1
        assert g[0] == 0
        assert g.data, list(range(10))
        assert g[8] == 8
        assert g[0] == 0

    def test_gen__index_2(self, g):
        assert g[0] == 0
        assert g[1] == 1
        assert g.data == list(range(10))

    def test_gen__index_error(self, g):
        assert g[0] == 0
        with pytest.raises(IndexError):
            g[11]
        assert list(iter(g)) == list(range(10))

    def test_gen__negative_index(self, g):
        assert g[-1] == 9
        assert g[-2] == 8
        assert g[-3] == 7
        assert g[-4] == 6
        assert g[-5] == 5
        assert g[5] == 5
        assert g.data == list(range(10))

        assert list(iter(g)) == list(range(10))

    def test_nonzero__does_not_consume_more_than_first_item(self):
        def build_generator():
            yield 1
            pytest.fail("generator should not consume past first item")
            yield 2

        g = regen(build_generator())
        assert bool(g)
        assert g[0] == 1

    def test_nonzero__empty_iter(self):
        assert not regen(iter([]))

    def test_deque(self):
        original_list = [42]
        d = collections.deque(original_list)
        # Confirm that concretising a `regen()` instance repeatedly for an
        # equality check always returns the original list
        g = regen(d)
        assert g == original_list
        assert g == original_list

    def test_repr(self):
        def die():
            raise AssertionError("Generator died")
            yield None

        # Confirm that `regen()` instances are not concretised when represented
        g = regen(die())
        assert "..." in repr(g)

    def test_partial_reconcretisation(self):
        class WeirdIterator():
            def __init__(self, iter_):
                self.iter_ = iter_
                self._errored = False

            def __iter__(self):
                yield from self.iter_
                if not self._errored:
                    try:
                        # This should stop the regen instance from marking
                        # itself as being done
                        raise AssertionError("Iterator errored")
                    finally:
                        self._errored = True

        original_list = list(range(42))
        g = regen(WeirdIterator(original_list))
        iter_g = iter(g)
        for e in original_list:
            assert e == next(iter_g)
        with pytest.raises(AssertionError, match="Iterator errored"):
            next(iter_g)
        # The following checks are for the known "misbehaviour"
        assert getattr(g, "_regen__done") is False
        # If the `regen()` instance doesn't think it's done then it'll dupe the
        # elements from the underlying iterator if it can be reused
        iter_g = iter(g)
        for e in original_list * 2:
            assert next(iter_g) == e
        with pytest.raises(StopIteration):
            next(iter_g)
        assert getattr(g, "_regen__done") is True
        # Finally we xfail this test to keep track of it
        raise pytest.xfail(reason="#6794")

    def test_length_hint_passthrough(self, g):
        assert g.__length_hint__() == 10

    def test_getitem_repeated(self, g):
        halfway_idx = g.__length_hint__() // 2
        assert g[halfway_idx] == halfway_idx
        # These are now concretised so they should be returned without any work
        assert g[halfway_idx] == halfway_idx
        for i in range(halfway_idx + 1):
            assert g[i] == i
        # This should only need to concretise one more element
        assert g[halfway_idx + 1] == halfway_idx + 1

    def test_done_does_not_lag(self, g):
        """
        Don't allow regen to return from `__iter__()` and check `__done`.
        """
        # The range we zip with here should ensure that the `regen.__iter__`
        # call never gets to return since we never attempt a failing `next()`
        len_g = g.__length_hint__()
        for i, __ in zip(range(len_g), g):
            assert getattr(g, "_regen__done") is (i == len_g - 1)
        # Just for sanity, check against a specific `bool` here
        assert getattr(g, "_regen__done") is True

    def test_lookahead_consume(self, subtests):
        """
        Confirm that regen looks ahead by a single item as expected.
        """
        def g():
            yield from ["foo", "bar"]
            raise pytest.fail("This should never be reached")

        with subtests.test(msg="bool does not overconsume"):
            assert bool(regen(g()))
        with subtests.test(msg="getitem 0th does not overconsume"):
            assert regen(g())[0] == "foo"
        with subtests.test(msg="single iter does not overconsume"):
            assert next(iter(regen(g()))) == "foo"

        class ExpectedException(BaseException):
            pass

        def g2():
            yield from ["foo", "bar"]
            raise ExpectedException()

        with subtests.test(msg="getitem 1th does overconsume"):
            r = regen(g2())
            with pytest.raises(ExpectedException):
                r[1]
            # Confirm that the item was concretised anyway
            assert r[1] == "bar"
        with subtests.test(msg="full iter does overconsume"):
            r = regen(g2())
            with pytest.raises(ExpectedException):
                for _ in r:
                    pass
            # Confirm that the items were concretised anyway
            assert r == ["foo", "bar"]
        with subtests.test(msg="data access does overconsume"):
            r = regen(g2())
            with pytest.raises(ExpectedException):
                r.data
            # Confirm that the items were concretised anyway
            assert r == ["foo", "bar"]


class test_head_from_fun:

    def test_from_cls(self):
        class X:
            def __call__(x, y, kwarg=1):
                pass

        g = head_from_fun(X())
        with pytest.raises(TypeError):
            g(1)
        g(1, 2)
        g(1, 2, kwarg=3)

    def test_from_fun(self):
        def f(x, y, kwarg=1):
            pass
        g = head_from_fun(f)
        with pytest.raises(TypeError):
            g(1)
        g(1, 2)
        g(1, 2, kwarg=3)

    def test_regression_3678(self):
        local = {}
        fun = ('def f(foo, *args, bar="", **kwargs):'
               '    return foo, args, bar')
        exec(fun, {}, local)

        g = head_from_fun(local['f'])
        g(1)
        g(1, 2, 3, 4, bar=100)
        with pytest.raises(TypeError):
            g(bar=100)

    def test_from_fun_with_hints(self):
        local = {}
        fun = ('def f_hints(x: int, y: int, kwarg: int=1):'
               '    pass')
        exec(fun, {}, local)
        f_hints = local['f_hints']

        g = head_from_fun(f_hints)
        with pytest.raises(TypeError):
            g(1)
        g(1, 2)
        g(1, 2, kwarg=3)

    def test_from_fun_forced_kwargs(self):
        local = {}
        fun = ('def f_kwargs(*, a, b="b", c=None):'
               '    return')
        exec(fun, {}, local)
        f_kwargs = local['f_kwargs']

        g = head_from_fun(f_kwargs)
        with pytest.raises(TypeError):
            g(1)

        g(a=1)
        g(a=1, b=2)
        g(a=1, b=2, c=3)

    def test_classmethod(self):
        class A:
            @classmethod
            def f(cls, x):
                return x

        fun = head_from_fun(A.f, bound=False)
        assert fun(A, 1) == 1

        fun = head_from_fun(A.f, bound=True)
        assert fun(1) == 1

    def test_kwonly_required_args(self):
        local = {}
        fun = ('def f_kwargs_required(*, a="a", b, c=None):'
               '    return')
        exec(fun, {}, local)
        f_kwargs_required = local['f_kwargs_required']
        g = head_from_fun(f_kwargs_required)

        with pytest.raises(TypeError):
            g(1)

        with pytest.raises(TypeError):
            g(a=1)

        with pytest.raises(TypeError):
            g(c=1)

        with pytest.raises(TypeError):
            g(a=2, c=1)

        g(b=3)

    @pytest.mark.skipif(sys.version_info < (3, 14), reason="PEP 649 deferred annotations require Python 3.14+")
    def test_type_checking_annotation(self):
        # Regression test for https://github.com/celery/celery/discussions/10099
        # On Python 3.14+, annotations are deferred (PEP 649). Functions with
        # annotations referencing TYPE_CHECKING-only types must not raise NameError.
        local = {}
        exec('def f(args: Sequence[str], x: int = 0): return args', {}, local)
        f = local['f']

        g = head_from_fun(f)
        with pytest.raises(TypeError):
            g()
        g(1)
        g(1, 2)


class test_fun_takes_argument:

    def test_starkwargs(self):
        assert fun_takes_argument('foo', lambda **kw: 1)

    def test_named(self):
        assert fun_takes_argument('foo', lambda a, foo, bar: 1)

        def fun(a, b, c, d):
            return 1

        assert fun_takes_argument('foo', fun, position=4)

    def test_starargs(self):
        assert fun_takes_argument('foo', lambda a, *args: 1)

    def test_does_not(self):
        assert not fun_takes_argument('foo', lambda a, bar, baz: 1)
        assert not fun_takes_argument('foo', lambda: 1)

        def fun(a, b, foo):
            return 1

        assert not fun_takes_argument('foo', fun, position=4)


@pytest.mark.parametrize('a,b,expected', [
    ((1, 2, 3), [4, 5], (1, 2, 3, 4, 5)),
    ((1, 2), [3, 4, 5], [1, 2, 3, 4, 5]),
    ([1, 2, 3], (4, 5), [1, 2, 3, 4, 5]),
    ([1, 2], (3, 4, 5), (1, 2, 3, 4, 5)),
])
def test_seq_concat_seq(a, b, expected):
    res = seq_concat_seq(a, b)
    assert type(res) is type(expected)
    assert res == expected


@pytest.mark.parametrize('a,b,expected', [
    ((1, 2, 3), 4, (1, 2, 3, 4)),
    ([1, 2, 3], 4, [1, 2, 3, 4]),
])
def test_seq_concat_item(a, b, expected):
    res = seq_concat_item(a, b)
    assert type(res) is type(expected)
    assert res == expected


class StarKwargsCallable:

    def __call__(self, **kwargs):
        return 1


class StarArgsStarKwargsCallable:

    def __call__(self, *args, **kwargs):
        return 1


class StarArgsCallable:

    def __call__(self, *args):
        return 1


class ArgsCallable:

    def __call__(self, a, b):
        return 1


class ArgsStarKwargsCallable:

    def __call__(self, a, b, **kwargs):
        return 1


class test_fun_accepts_kwargs:

    @pytest.mark.parametrize('fun', [
        lambda a, b, **kwargs: 1,
        lambda *args, **kwargs: 1,
        lambda foo=1, **kwargs: 1,
        StarKwargsCallable(),
        StarArgsStarKwargsCallable(),
        ArgsStarKwargsCallable(),
    ])
    def test_accepts(self, fun):
        assert fun_accepts_kwargs(fun)

    @pytest.mark.parametrize('fun', [
        lambda a: 1,
        lambda a, b: 1,
        lambda *args: 1,
        lambda a, kw1=1, kw2=2: 1,
        StarArgsCallable(),
        ArgsCallable(),
    ])
    def test_rejects(self, fun):
        assert not fun_accepts_kwargs(fun)

    @pytest.mark.skipif(sys.version_info < (3, 14), reason="PEP 649 deferred annotations require Python 3.14+")
    def test_type_checking_annotation(self):
        # Regression test for https://github.com/celery/celery/discussions/10099
        # On Python 3.14+, annotations are deferred (PEP 649). Calling
        # fun_accepts_kwargs on a function whose annotations reference
        # TYPE_CHECKING-only types must not raise NameError.
        #
        # This reproduces the failure seen with on_after_finalize.connect:
        #   def setup_periodic_tasks(sender: Celery, **kwargs: object) -> None: ...
        # where 'Celery' is only imported under TYPE_CHECKING.
        local = {}
        exec('def f(sender: Celery, **kwargs: object) -> None: pass', {}, local)
        f = local['f']
        assert fun_accepts_kwargs(f) is True

        exec('def g(sender: Celery) -> None: pass', {}, local)
        g = local['g']
        assert fun_accepts_kwargs(g) is False


@pytest.mark.parametrize('value,expected', [
    (5, True),
    (5.0, True),
    (0, True),
    (0.0, True),
    (True, False),
    ('value', False),
    ('5', False),
    ('5.0', False),
    (None, False),
])
def test_is_numeric_value(value, expected):
    res = is_numeric_value(value)
    assert type(res) is type(expected)
    assert res == expected



import socket
import time

import pytest
from kombu import Connection

from celery import Celery


def wait_for_port(host, port, timeout=60.0):
    """Wait for a port to become available."""
    start = time.time()
    while time.time() - start < timeout:
        try:
            with socket.create_connection((host, port), timeout=2):
                return
        except OSError:
            time.sleep(1)
    raise TimeoutError(f"Timed out waiting for {host}:{port}")


@pytest.fixture()
def redis():
    """Fixture to provide Redis hostname and port."""
    return {"hostname": "redis", "port": 6379}


@pytest.fixture()
def app(rabbitmq, redis):
    wait_for_port(rabbitmq.hostname, rabbitmq.ports[5672])
    wait_for_port(redis["hostname"], redis["port"])

    return Celery(
        "test_app",
        broker=f"pyamqp://guest:guest@{rabbitmq.hostname}:{rabbitmq.ports[5672]}/",
        backend=f"redis://{redis['hostname']}:{redis['port']}/0",
        include=["t.integration.test_rabbitmq_default_queue_type_fallback"],
    )


@pytest.fixture()
def ping(app):
    @app.task(name="ping")
    def ping():
        return "pong"
    return ping


@pytest.mark.amqp
@pytest.mark.timeout(60)
@pytest.mark.xfail(
    reason=(
        "Celery does not respect task_default_exchange_type/queue_type "
        "when using implicit routing to the 'celery' queue. It creates "
        "a classic queue and direct exchange instead."
    ),
    strict=True,
)
def test_fallback_to_classic_queue_and_direct_exchange(app, ping):
    from celery.contrib.testing.worker import start_worker

    # Start worker and submit task
    with start_worker(app, queues=["celery"], loglevel="info", perform_ping_check=False):
        result = ping.delay()
        assert result.get(timeout=10) == "pong"

        exchange_type = None
        start_time = time.time()
        timeout = 10  # Maximum wait time in seconds

        while time.time() - start_time < timeout:
            with Connection(app.conf.broker_url) as conn:
                with conn.channel() as channel:
                    try:
                        response = channel.exchange_declare("celery", passive=True)
                        exchange_type = response['type']
                        break
                    except Exception:
                        time.sleep(0.5)

        if exchange_type is None:
            exchange_type = "error: Exchange declaration timed out"
        assert exchange_type != "direct", (
            "Expected Celery to honor task_default_exchange_type, "
            f"but got: {exchange_type}"
        )



from __future__ import annotations

import json

import pytest
from pytest_celery import RESULT_TIMEOUT, CeleryTestSetup, CeleryTestWorker, CeleryWorkerCluster

from celery.canvas import StampingVisitor, chain
from t.integration.tasks import StampOnReplace, identity, replace_with_stamped_task


def get_hybrid_clusters_matrix() -> list[list[str]]:
    """Returns a matrix of hybrid worker clusters

    Each item in the matrix is a list of workers to be used in the cluster
    and each cluster will be tested separately (with parallel support)
    """

    return [
        # Dev worker only
        ["celery_setup_worker"],
        # Legacy (Celery 4) worker only
        ["celery_legacy_worker"],
        # Both dev and legacy workers
        ["celery_setup_worker", "celery_legacy_worker"],
        # Dev worker and last official Celery release worker
        ["celery_setup_worker", "celery_latest_worker"],
        # Dev worker and legacy worker and last official Celery release worker
        ["celery_setup_worker", "celery_latest_worker", "celery_legacy_worker"],
    ]


@pytest.fixture(params=get_hybrid_clusters_matrix())
def celery_worker_cluster(request: pytest.FixtureRequest) -> CeleryWorkerCluster:
    nodes: tuple[CeleryTestWorker] = [
        request.getfixturevalue(worker) for worker in request.param
    ]
    cluster = CeleryWorkerCluster(*nodes)
    yield cluster
    cluster.teardown()


class test_stamping_hybrid_worker_cluster:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        stamp = {"stamp": 42}

        class CustomStampingVisitor(StampingVisitor):
            def on_signature(self, sig, **headers) -> dict:
                return stamp.copy()

        worker: CeleryTestWorker
        for worker in celery_setup.worker_cluster:
            queue = worker.worker_queue
            stamped_task = identity.si(123)
            stamped_task.stamp(visitor=CustomStampingVisitor())
            assert stamped_task.apply_async(queue=queue).get(timeout=RESULT_TIMEOUT)
            assert worker.logs().count(json.dumps(stamp, indent=4, sort_keys=True))

    def test_sanity_worker_hop(self, celery_setup: CeleryTestSetup):
        if len(celery_setup.worker_cluster) < 2:
            pytest.skip("Not enough workers in cluster")

        stamp = {"stamp": 42}

        class CustomStampingVisitor(StampingVisitor):
            def on_signature(self, sig, **headers) -> dict:
                return stamp.copy()

        w1: CeleryTestWorker = celery_setup.worker_cluster[0]
        w2: CeleryTestWorker = celery_setup.worker_cluster[1]
        stamped_task = chain(
            identity.si(4).set(queue=w1.worker_queue),
            identity.si(2).set(queue=w2.worker_queue),
        )
        stamped_task.stamp(visitor=CustomStampingVisitor())
        stamped_task.apply_async().get(timeout=RESULT_TIMEOUT)

        stamp = json.dumps(stamp, indent=4)
        worker: CeleryTestWorker
        for worker in (w1, w2):
            assert worker.logs().count(stamp)

    def test_multiple_stamps_multiple_workers(self, celery_setup: CeleryTestSetup):
        if len(celery_setup.worker_cluster) < 2:
            pytest.skip("Not enough workers in cluster")

        stamp = {"stamp": 420}
        stamp1 = {**stamp, "stamp1": 4}
        stamp2 = {**stamp, "stamp2": 2}

        w1: CeleryTestWorker = celery_setup.worker_cluster[0]
        w2: CeleryTestWorker = celery_setup.worker_cluster[1]
        stamped_task = chain(
            identity.si(4).set(queue=w1.worker_queue).stamp(stamp1=stamp1["stamp1"]),
            identity.si(2).set(queue=w2.worker_queue).stamp(stamp2=stamp2["stamp2"]),
        )
        stamped_task.stamp(stamp=stamp["stamp"])
        stamped_task.apply_async().get(timeout=RESULT_TIMEOUT)

        stamp1 = json.dumps(stamp1, indent=4)
        stamp2 = json.dumps(stamp2, indent=4)

        assert w1.logs().count(stamp1)
        assert w1.logs().count(stamp2) == 0

        assert w2.logs().count(stamp1) == 0
        assert w2.logs().count(stamp2)

    def test_stamping_on_replace_with_legacy_worker_in_cluster(
        self,
        celery_setup: CeleryTestSetup,
        dev_worker: CeleryTestWorker,
        legacy_worker: CeleryTestWorker,
    ):
        if len(celery_setup.worker_cluster) < 2:
            pytest.skip("Not enough workers in cluster")

        if not dev_worker:
            pytest.skip("Dev worker not in cluster")

        if not legacy_worker:
            pytest.skip("Legacy worker not in cluster")

        stamp = {"stamp": "Only for dev worker tasks"}
        stamp1 = {**StampOnReplace.stamp, "stamp1": "1) Only for legacy worker tasks"}
        stamp2 = {**StampOnReplace.stamp, "stamp2": "2) Only for legacy worker tasks"}

        replaced_sig1 = (
            identity.si(4)
            .set(queue=legacy_worker.worker_queue)
            .stamp(stamp1=stamp1["stamp1"])
        )
        replaced_sig2 = (
            identity.si(2)
            .set(queue=legacy_worker.worker_queue)
            .stamp(stamp2=stamp2["stamp2"])
        )

        stamped_task = chain(
            replace_with_stamped_task.si(replace_with=replaced_sig1).set(
                queue=dev_worker.worker_queue
            ),
            replace_with_stamped_task.si(replace_with=replaced_sig2).set(
                queue=dev_worker.worker_queue
            ),
        )
        stamped_task.stamp(stamp=stamp["stamp"])
        stamped_task.apply_async().get(timeout=RESULT_TIMEOUT)

        stamp = json.dumps(stamp, indent=4)
        stamp1 = json.dumps(stamp1, indent=4)
        stamp2 = json.dumps(stamp2, indent=4)

        assert dev_worker.logs().count(stamp)
        assert dev_worker.logs().count(stamp1) == 0
        assert dev_worker.logs().count(stamp2) == 0

        assert legacy_worker.logs().count(stamp) == 0
        assert legacy_worker.logs().count(stamp1)
        assert legacy_worker.logs().count(stamp2)



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from unittest.mock import Mock

import pytest
from pytest_celery import CeleryTestSetup, CeleryTestWorker, CeleryWorkerCluster

from celery import Celery
from celery.app.base import set_default_app
from celery.signals import after_task_publish
from t.integration.tasks import identity


@pytest.fixture(
    params=[
        # Single worker
        ["celery_setup_worker"],
        # Workers cluster (same queue)
        ["celery_setup_worker", "celery_alt_dev_worker"],
    ]
)
def celery_worker_cluster(request: pytest.FixtureRequest) -> CeleryWorkerCluster:
    nodes: tuple[CeleryTestWorker] = [
        request.getfixturevalue(worker) for worker in request.param
    ]
    cluster = CeleryWorkerCluster(*nodes)
    yield cluster
    cluster.teardown()


class test_thread_safety:
    @pytest.fixture
    def default_worker_app(self, default_worker_app: Celery) -> Celery:
        app = default_worker_app
        app.conf.broker_pool_limit = 42
        return app

    @pytest.mark.parametrize(
        "threads_count",
        [
            # Single
            1,
            # Multiple
            2,
            # Many
            42,
        ],
    )
    def test_multithread_task_publish(
        self,
        celery_setup: CeleryTestSetup,
        threads_count: int,
    ):
        signal_was_called = Mock()

        @after_task_publish.connect
        def after_task_publish_handler(*args, **kwargs):
            signal_was_called(True)

        def thread_worker():
            set_default_app(celery_setup.app)
            identity.si("Published from thread").apply_async(
                queue=celery_setup.worker.worker_queue
            )

        executor = ThreadPoolExecutor(threads_count)

        with executor:
            for _ in range(threads_count):
                executor.submit(thread_worker)

        assert signal_was_called.call_count == threads_count



from datetime import datetime, timedelta, timezone

import pytest

from celery.exceptions import CPendingDeprecationWarning
from celery.utils.iso8601 import parse_iso8601


def test_parse_iso8601_utc():
    dt = parse_iso8601("2023-10-26T10:30:00Z")
    assert dt == datetime(2023, 10, 26, 10, 30, 0, tzinfo=timezone.utc)


def test_parse_iso8601_positive_offset():
    dt = parse_iso8601("2023-10-26T10:30:00+05:30")
    expected_tz = timezone(timedelta(hours=5, minutes=30))
    assert dt == datetime(2023, 10, 26, 10, 30, 0, tzinfo=expected_tz)


def test_parse_iso8601_negative_offset():
    dt = parse_iso8601("2023-10-26T10:30:00-08:00")
    expected_tz = timezone(timedelta(hours=-8))
    assert dt == datetime(2023, 10, 26, 10, 30, 0, tzinfo=expected_tz)


def test_parse_iso8601_with_microseconds():
    dt = parse_iso8601("2023-10-26T10:30:00.123456Z")
    assert dt == datetime(2023, 10, 26, 10, 30, 0, 123456, tzinfo=timezone.utc)


def test_parse_iso8601_date_only():
    dt = parse_iso8601("2023-10-26")
    assert dt == datetime(2023, 10, 26, 0, 0, 0)  # Expects naive datetime


def test_parse_iso8601_date_hour_minute_only():
    # The regex uses '.' as a separator, often 'T' is used.
    # Let's test with 'T' as it's common in ISO8601.
    dt = parse_iso8601("2023-10-26T10:30")
    assert dt == datetime(2023, 10, 26, 10, 30, 0)  # Expects naive datetime


def test_parse_iso8601_invalid_string():
    with pytest.raises(ValueError, match="unable to parse date string"):
        parse_iso8601("invalid-date-string")


def test_parse_iso8601_malformed_strings():
    # These strings match the regex but have invalid date/time component values
    invalid_component_strings = [
        "2023-13-01T00:00:00Z",  # Invalid month
        "2023-12-32T00:00:00Z",  # Invalid day
        "2023-12-01T25:00:00Z",  # Invalid hour
        "2023-12-01T00:60:00Z",  # Invalid minute
        "2023-12-01T00:00:60Z",  # Invalid second
    ]
    for s in invalid_component_strings:
        # For these, the error comes from datetime constructor
        with pytest.raises(ValueError):
            parse_iso8601(s)

    # This string has a timezone format that is ignored by the parser, resulting in a naive datetime
    ignored_tz_string = "2023-10-26T10:30:00+05:AA"
    dt_ignored_tz = parse_iso8601(ignored_tz_string)
    assert dt_ignored_tz == datetime(2023, 10, 26, 10, 30, 0)
    assert dt_ignored_tz.tzinfo is None

    # This string does not match the main ISO8601_REGEX pattern correctly, leading to None groups
    unparseable_string = "20231026T103000Z"
    with pytest.raises(TypeError):  # Expects TypeError due to int(None)
        parse_iso8601(unparseable_string)


def test_parse_iso8601_deprecation_warning():
    with pytest.warns(CPendingDeprecationWarning, match="parse_iso8601 is scheduled for deprecation"):
        parse_iso8601("2023-10-26T10:30:00Z")
